@@ -5,7 +5,7 @@
 \*/
 
 /*
- * $Id: docode.h,v 1.7 1998/05/25 10:38:45 hubbe Exp $
+ * $Id: docode.h,v 1.11 2001/02/25 17:55:08 hubbe Exp $
  */
 #ifndef DOCODE_H
 #define DOCODE_H
@@ -15,13 +15,15 @@
 #define DO_POP 4
 #define DO_INDIRECT 8
 #define DO_LVALUE_IF_POSSIBLE 16
+#define DO_NOT_COPY_TOPLEVEL 32
 
 #define WANT_LVALUE (DO_LVALUE | DO_INDIRECT)
 
 extern int store_linenumbers;
 
-#define emit(X,Y) insert_opcode((X),(Y),lex.current_line, lex.current_file)
-#define emit2(X) insert_opcode2((X),lex.current_line, lex.current_file)
+#define emit0(X)     insert_opcode0((X),lex.current_line, lex.current_file)
+#define emit1(X,Y)   insert_opcode1((X),(Y),lex.current_line, lex.current_file)
+#define emit2(X,Y,Z) insert_opcode2((X),(Y),(Z),lex.current_line, lex.current_file)
 
 /* Prototypes begin here */
 void upd_int(int offset, INT32 tmp);
@@ -32,7 +34,7 @@ INT32 pop_address(void);
 int alloc_label(void);
 int do_jump(int token,INT32 lbl);
 void do_pop(int x);
-int do_docode(node *n,INT16 flags);
+int do_docode(node *n, INT16 flags);
 void do_cond_jump(node *n, int label, int iftrue, int flags);
 void do_code_block(node *n);
 int docode(node *n);
