@@ -1,5 +1,5 @@
 /*
- * $Id: system.c,v 1.100 2001/04/09 16:51:05 jonasw Exp $
+ * $Id: system.c,v 1.105 2003/10/17 15:39:50 mast Exp $
  *
  * System-call module for Pike
  *
@@ -15,7 +15,7 @@
 #include "system_machine.h"
 #include "system.h"
 
-RCSID("$Id: system.c,v 1.100 2001/04/09 16:51:05 jonasw Exp $");
+RCSID("$Id: system.c,v 1.105 2003/10/17 15:39:50 mast Exp $");
 #ifdef HAVE_WINSOCK_H
 #include <winsock.h>
 #endif
@@ -92,6 +92,10 @@ RCSID("$Id: system.c,v 1.100 2001/04/09 16:51:05 jonasw Exp $");
 #include <sys/id.h>
 #endif /* HAVE_SYS_ID_H */
 
+#ifdef HAVE_SYS_PRCTL_H
+#include <sys/prctl.h>
+#endif /* HAVE_SYS_PRCTL_H */
+
 #ifdef HAVE_UTIME_H
 #include <utime.h>
 #endif
@@ -806,6 +810,51 @@ void f_setsid(INT32 args)
 }
 #endif
 
+#if defined(HAVE_PRCTL) && defined(PR_SET_DUMPABLE)
+/*! @decl int(0..1) dumpable(int(0..1)|void val)
+ *!   Get and/or set whether this process should be able to
+ *!   dump core.
+ *!
+ *! @param val
+ *!   Optional argument to set the core dumping state.
+ *!   @int
+ *!     @value 0
+ *!       Disable core dumping for this process.
+ *!     @value 1
+ *!       Enable core dumping for this process.
+ *!   @endint
+ *!
+ *! @returns
+ *!   Returns @tt{1@} if this process currently is capable of dumping core,
+ *!   and @tt{0@} (zero) if not.
+ *!
+ *! @note
+ *!   This function is currently only available on some versions of Linux.
+ */ 
+void f_dumpable(INT32 args)
+{
+  int current = prctl(PR_GET_DUMPABLE);
+
+  if (current == -1) {
+    int err = errno;
+    Pike_error("Failed to get dumpable state. errno:%d\n", err);
+  }
+  if (args) {
+    INT_TYPE val;
+    get_all_args("dumpable", args, "%i", &val);
+    if (val & ~1) {
+      SIMPLE_BAD_ARG_ERROR("dumpable", 1, "int(0..1)");
+    }
+    if (prctl(PR_SET_DUMPABLE, val) == -1) {
+      int err = errno;
+      Pike_error("Failed to set dumpable state to %d. errno:%d\n", val, err);
+    }
+  }
+  pop_n_elems(args);
+  push_int(current);
+}
+#endif
+
 #ifdef HAVE_SETRESUID
 void f_setresuid(INT32 args)
 {
@@ -1830,8 +1879,7 @@ static void f_get_netinfo_property(INT32 args)
   
   get_all_args("get_netinfo_property", args, "%s%s%s",
 	       &domain_str, &path_str, &prop_str);
-  pop_n_elems(args);
-  
+
   /* open domain */
   num_replies = 0;
   res = ni_open(NULL, domain_str, &dom);
@@ -1858,6 +1906,7 @@ static void f_get_netinfo_property(INT32 args)
     f_aggregate(num_replies);
   else
     push_int(0);
+  stack_pop_n_elems_keep_top (args);
 }
 #endif
 
@@ -1867,6 +1916,10 @@ static void f_get_netinfo_property(INT32 args)
 
 void pike_module_init(void)
 {
+#ifdef GETHOSTBYNAME_MUTEX_EXISTS
+  mt_init(&gethostbyname_mutex);
+#endif
+
   /*
    * From this file:
    */
@@ -2048,6 +2101,11 @@ void pike_module_init(void)
 		OPT_SIDE_EFFECT);
 #endif
 
+#if defined(HAVE_PRCTL) && defined(PR_SET_DUMPABLE)
+  ADD_FUNCTION2("dumpable", f_dumpable, tFunc(tOr(tInt01, tVoid), tInt01),
+		0, OPT_SIDE_EFFECT);
+#endif
+
 #ifdef HAVE_GETRLIMIT
   ADD_FUNCTION2("getrlimit", f_getrlimit, tFunc(tString, tArr(tInt)), 
 		0, OPT_EXTERNAL_DEPEND);
@@ -2104,15 +2162,15 @@ void pike_module_init(void)
 #ifdef HAVE_SYSLOG
   
 /* function(string,int,int:void) */
-  ADD_EFUN("openlog", f_openlog,tFunc(tStr tInt tInt,tVoid), 0);
+  ADD_EFUN("openlog", f_openlog,tFunc(tStr tInt tInt,tVoid), OPT_SIDE_EFFECT);
   ADD_FUNCTION("openlog", f_openlog,tFunc(tStr tInt tInt,tVoid), 0);
   
 /* function(int,string:void) */
-  ADD_EFUN("syslog", f_syslog,tFunc(tInt tStr,tVoid), 0);
+  ADD_EFUN("syslog", f_syslog,tFunc(tInt tStr,tVoid), OPT_SIDE_EFFECT);
   ADD_FUNCTION("syslog", f_syslog,tFunc(tInt tStr,tVoid), 0);
   
 /* function(:void) */
-  ADD_EFUN("closelog", f_closelog,tFunc(tNone,tVoid), 0);
+  ADD_EFUN("closelog", f_closelog,tFunc(tNone,tVoid), OPT_SIDE_EFFECT);
   ADD_FUNCTION("closelog", f_closelog,tFunc(tNone,tVoid), 0);
 #endif /* HAVE_SYSLOG */
 
@@ -2153,4 +2211,8 @@ void pike_module_exit(void)
     exit_nt_system_calls();
   }
 #endif
+
+#ifdef GETHOSTBYNAME_MUTEX_EXISTS
+  mt_destroy(&gethostbyname_mutex);
+#endif
 }
