@@ -62,7 +62,7 @@ static int pike_isnan(double x)
 #endif /* HAVE__ISNAN */
 #endif /* HAVE_ISNAN */
 
-RCSID("$Id: svalue.c,v 1.106 2001/11/14 10:51:17 grubba Exp $");
+RCSID("$Id: svalue.c,v 1.107 2003/09/08 15:27:58 mast Exp $");
 
 struct svalue dest_ob_zero = { T_INT, 0 };
 
@@ -1516,9 +1516,8 @@ void gc_check_weak_short_svalue(union anything *u, TYPE_T type)
       }
 
 #define GC_DO_MARK(U, TN)						\
-      enqueue(&gc_mark_queue,						\
-	      (queue_call) PIKE_CONCAT3(gc_mark_, TN, _as_referenced),	\
-	      U.TN)
+      gc_mark_enqueue((queue_call) PIKE_CONCAT3(gc_mark_, TN, _as_referenced),	\
+		      U.TN)
 
 #define GC_DONT_MARK(U, TN) do {} while (0)
 
