@@ -1,18 +1,31 @@
 /*
- * $Id: threads.h,v 1.64 1999/05/14 22:30:29 hubbe Exp $
+ * $Id: threads.h,v 1.120 2003/12/13 20:37:53 jonasw Exp $
  */
 #ifndef THREADS_H
 #define THREADS_H
 
 #include "machine.h"
-#include "interpret.h"
 #include "object.h"
-#include "error.h"
+#include "pike_error.h"
+#include "interpret.h"
+#include "main.h"
+
+/* Needed for the sigset_t typedef, which is needed for
+ * the pthread_sigsetmask() prototype on Solaris 2.x.
+ */
+#include <signal.h>
+
 #ifdef HAVE_SYS_TYPES_H
 /* Needed for pthread_t on OSF/1 */
 #include <sys/types.h>
 #endif /* HAVE_SYS_TYPES_H */
-#ifdef _REENTRANT
+#ifdef PIKE_THREADS
+
+/* The fp macro conflicts with Solaris's <pthread.h>. */
+#ifdef fp
+#undef fp
+#define FRAMEPOINTER_WAS_DEFINED
+#endif /* fp */
 
 /*
  * Decide which type of threads to use
@@ -33,11 +46,23 @@
 #endif /* _UNIX_THREADS */
 
 #ifdef _MIT_POSIX_THREADS
-#ifdef HAVE_PTHREAD_H
 #define POSIX_THREADS
 #include <pthread.h>
-#undef HAVE_PTHREAD_H
+
+#ifdef HAVE_MACH_TASK_INFO_H
+#include <mach/task_info.h>
 #endif
+#ifdef HAVE_MACH_TASK_H
+#include <mach/task.h>
+#endif
+#ifdef HAVE_MACH_MACH_INIT_H
+#include <mach/mach_init.h>
+#endif
+
+/* AIX is *STUPID* - Hubbe */
+#undef func_data
+
+#undef HAVE_PTHREAD_H
 #endif /* _MIT_POSIX_THREADS */
 
 #ifdef _SGI_SPROC_THREADS
@@ -47,22 +72,89 @@
 #endif /* _SGI_SPROC_THREADS */
 
 
-extern int num_threads;
-extern int live_threads;
-struct object;
-extern size_t thread_stack_size;
-extern struct object *thread_id;
 
-#define DEFINE_MUTEX(X) MUTEX_T X
+/* Restore the fp macro. */
+#ifdef FRAMEPOINTER_WAS_DEFINED
+#define fp Pike_fp
+#undef FRAMEPOINTER_WAS_DEFINED
+#endif /* FRAMEPOINTER_WAS_DEFINED */
+
+
+extern int num_threads;
+PMOD_EXPORT extern int live_threads;
+struct object;
+PMOD_EXPORT extern size_t thread_stack_size;
+
+PMOD_EXPORT void thread_low_error (int errcode);
+
+#define LOW_THREAD_CHECK_NONZERO_ERROR(CALL) do {			\
+    int thread_errcode_ = (CALL);					\
+    if (thread_errcode_) thread_low_error (thread_errcode_);		\
+  } while (0)
+
+#define DEFINE_MUTEX(X) PIKE_MUTEX_T X
 
 
 #ifdef POSIX_THREADS
+
+#ifdef HAVE_PTHREAD_ATFORK
+#define th_atfork(X,Y,Z) pthread_atfork((X),(Y),(Z))
+#define th_atfork_prepare()
+#define th_atfork_parent()
+#define th_atfork_child()
+#else
+int th_atfork(void (*)(void),void (*)(void),void (*)(void));
+void th_atfork_prepare(void);
+void th_atfork_parent(void);
+void th_atfork_child(void);
+#endif
+
 #define THREAD_T pthread_t
-#define MUTEX_T pthread_mutex_t
+#define PIKE_MUTEX_T pthread_mutex_t
+
+#if !defined(HAVE_PTHREAD_MUTEX_ERRORCHECK_NP) && defined(HAVE_PTHREAD_MUTEX_ERRORCHECK)
+#define PTHREAD_MUTEX_ERRORCHECK_NP PTHREAD_MUTEX_ERRORCHECK
+#define HAVE_PTHREAD_MUTEX_ERRORCHECK_NP
+#endif
+
+#ifdef HAVE_PTHREAD_MUTEX_ERRORCHECK_NP
+#define mt_init(X) do {							\
+    if (debug_options & ERRORCHECK_MUTEXES) {				\
+      pthread_mutexattr_t attr;						\
+      pthread_mutexattr_init(&attr);					\
+      pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_ERRORCHECK_NP);	\
+      pthread_mutex_init((X), &attr);					\
+    }									\
+    else								\
+      pthread_mutex_init((X),0);					\
+  } while (0)
+#define DO_IF_PIKE_MUTEX_ERRORCHECK(X) X
+#else
 #define mt_init(X) pthread_mutex_init((X),0)
-#define mt_lock(X) pthread_mutex_lock(X)
+#define DO_IF_PIKE_MUTEX_ERRORCHECK(X)
+#endif
+
+#if !defined(HAVE_PTHREAD_MUTEX_RECURSIVE_NP) && defined(HAVE_PTHREAD_MUTEX_RECURSIVE)
+#define PTHREAD_MUTEX_RECURSIVE_NP PTHREAD_MUTEX_RECURSIVE
+#define HAVE_PTHREAD_MUTEX_RECURSIVE_NP
+#endif
+
+#ifdef HAVE_PTHREAD_MUTEX_RECURSIVE_NP
+#define mt_init_recursive(X)						\
+    do{ \
+      pthread_mutexattr_t attr;					\
+      pthread_mutexattr_init(&attr);					\
+      pthread_mutexattr_settype(					\
+	&attr,								\
+	PTHREAD_MUTEX_RECURSIVE_NP					\
+	DO_IF_PIKE_MUTEX_ERRORCHECK (| PTHREAD_MUTEX_ERRORCHECK_NP));	\
+      pthread_mutex_init((X), &attr);					\
+    }while(0)
+#endif
+
+#define mt_lock(X) LOW_THREAD_CHECK_NONZERO_ERROR (pthread_mutex_lock(X))
 #define mt_trylock(X) pthread_mutex_trylock(X)
-#define mt_unlock(X) pthread_mutex_unlock(X)
+#define mt_unlock(X) LOW_THREAD_CHECK_NONZERO_ERROR (pthread_mutex_unlock(X))
 #define mt_destroy(X) pthread_mutex_destroy(X)
 
 /* SIGH! No setconcurrency in posix threads. This is more or less
@@ -74,7 +166,10 @@ extern struct object *thread_id;
 #ifdef HAVE_PTHREAD_YIELD
 #define th_yield()	pthread_yield()
 #else
-#define th_yield()
+#ifdef HAVE_PTHREAD_YIELD_NP
+/* Some pthread libs define yield as non-portable function. */
+#define th_yield()	pthread_yield_np()
+#endif /* HAVE_PTHREAD_YIELD_NP */
 #endif /* HAVE_PTHREAD_YIELD */
 extern pthread_attr_t pattr;
 extern pthread_attr_t small_pattr;
@@ -83,7 +178,19 @@ extern pthread_attr_t small_pattr;
 #define th_create_small(ID,fun,arg) pthread_create(ID,&small_pattr,fun,arg)
 #define th_exit(foo) pthread_exit(foo)
 #define th_self() pthread_self()
+
+#define TH_KEY_T pthread_key_t
+#define th_key_create pthread_key_create
+#define th_setspecific pthread_setspecific
+#define th_getpecific pthread_getspecific
+
+
+#ifdef HAVE_PTHREAD_KILL
 #define th_kill(ID,sig) pthread_kill((ID),(sig))
+#else /* !HAVE_PTHREAD_KILL */
+/* MacOS X (aka Darwin) prior to 10.2 doesn't have pthread_kill. */
+#define th_kill(ID,sig)
+#endif /* HAVE_PTHREAD_KILL */
 #define th_join(ID,res) pthread_join((ID),(res))
 #ifdef HAVE_PTHREAD_COND_INIT
 #define COND_T pthread_cond_t
@@ -116,17 +223,17 @@ extern pthread_attr_t small_pattr;
 #ifdef UNIX_THREADS
 #define THREAD_T thread_t
 #define PTHREAD_MUTEX_INITIALIZER DEFAULTMUTEX
-#define MUTEX_T mutex_t
+#define PIKE_MUTEX_T mutex_t
 #define mt_init(X) mutex_init((X),USYNC_THREAD,0)
-#define mt_lock(X) mutex_lock(X)
+#define mt_lock(X) LOW_THREAD_CHECK_NONZERO_ERROR (mutex_lock(X))
 #define mt_trylock(X) mutex_trylock(X)
-#define mt_unlock(X) mutex_unlock(X)
+#define mt_unlock(X) LOW_THREAD_CHECK_NONZERO_ERROR (mutex_unlock(X))
 #define mt_destroy(X) mutex_destroy(X)
 
 #define th_setconcurrency(X) thr_setconcurrency(X)
 
 #define th_create(ID,fun,arg) thr_create(NULL,thread_stack_size,fun,arg,THR_DAEMON|THR_DETACHED,ID)
-#define th_create_small(ID,fun,arg) thr_create(NULL,32768,fun,arg,THR_DAEMON|THR_DETACHED,ID)
+#define th_create_small(ID,fun,arg) thr_create(NULL,8192*sizeof(char *),fun,arg,THR_DAEMON|THR_DETACHED,ID)
 #define th_exit(foo) thr_exit(foo)
 #define th_self() thr_self()
 #define th_kill(ID,sig) thr_kill((ID),(sig))
@@ -150,7 +257,7 @@ extern pthread_attr_t small_pattr;
  */
 #define THREAD_T	int
 
-#define MUTEX_T		ulock_t
+#define PIKE_MUTEX_T		ulock_t
 #define mt_init(X)	(usinitlock(((*X) = usnewlock(/*********/))))
 #define mt_lock(X)	ussetlock(*X)
 #define mt_unlock(X)	usunsetlock(*X)
@@ -178,6 +285,10 @@ extern pthread_attr_t small_pattr;
 #ifdef NT_THREADS
 #include <process.h>
 #include <windows.h>
+ 
+#define LOW_THREAD_CHECK_ZERO_ERROR(CALL) do {				\
+    if (!(CALL)) thread_low_error (GetLastError());			\
+  } while (0)
 
 #define THREAD_T unsigned
 #define th_setconcurrency(X)
@@ -191,25 +302,31 @@ extern pthread_attr_t small_pattr;
 #define th_destroy(X)
 #define th_yield() Sleep(0)
 #define th_equal(X,Y) ((X)==(Y))
-#define th_hash(X) ((unsigned INT32)(X))
+#define th_hash(X) (X)
 
-#define MUTEX_T HANDLE
+#define PIKE_MUTEX_T HANDLE
 #define mt_init(X) CheckValidHandle((*(X)=CreateMutex(NULL, 0, NULL)))
-#define mt_lock(X) (CheckValidHandle(*(X)),WaitForSingleObject(*(X), INFINITE))
-#define mt_trylock(X) (CheckValidHandle(*(X)),WaitForSingleObject(*(X), 0))
-#define mt_unlock(X) (CheckValidHandle(*(X)),ReleaseMutex(*(X)))
-#define mt_destroy(X) (CheckValidHandle(*(X)),CloseHandle(*(X)))
+#define mt_lock(X)							\
+  LOW_THREAD_CHECK_ZERO_ERROR (						\
+    WaitForSingleObject(CheckValidHandle(*(X)), INFINITE) == WAIT_OBJECT_0)
+#define mt_trylock(X)							\
+  LOW_THREAD_CHECK_ZERO_ERROR (						\
+    WaitForSingleObject(CheckValidHandle(*(X)), 0) != WAIT_FAILED)
+#define mt_unlock(X) LOW_THREAD_CHECK_ZERO_ERROR (ReleaseMutex(CheckValidHandle(*(X))))
+#define mt_destroy(X) CloseHandle(CheckValidHandle(*(X)))
 
 #define EVENT_T HANDLE
 #define event_init(X) CheckValidHandle(*(X)=CreateEvent(NULL, 1, 0, NULL))
-#define event_signal(X) (CheckValidHandle(*(X)),SetEvent(*(X)))
-#define event_destroy(X) (CheckValidHandle(*(X)),CloseHandle(*(X)))
-#define event_wait(X) (CheckValidHandle(*(X)),WaitForSingleObject(*(X), INFINITE))
+#define event_signal(X) LOW_THREAD_CHECK_ZERO_ERROR (SetEvent(CheckValidHandle(*(X))))
+#define event_destroy(X) CloseHandle(CheckValidHandle(*(X)))
+#define event_wait(X)							\
+  LOW_THREAD_CHECK_ZERO_ERROR (						\
+    WaitForSingleObject(CheckValidHandle(*(X)), INFINITE) == WAIT_OBJECT_0)
 
 #endif
 
 
-#if !defined(COND_T) && defined(EVENT_T) && defined(MUTEX_T)
+#if !defined(COND_T) && defined(EVENT_T) && defined(PIKE_MUTEX_T)
 
 #define SIMULATE_COND_WITH_EVENT
 
@@ -221,37 +338,73 @@ struct cond_t_queue
 
 typedef struct cond_t_s
 {
-  MUTEX_T lock;
+  PIKE_MUTEX_T lock;
   struct cond_t_queue *head, *tail;
 } COND_T;
 
 #define COND_T struct cond_t_s
 
-#define co_init(X) do { mt_init(& (X)->lock), (X)->head=(X)->tail=0; }while(0)
+#define co_init(X) do { mt_init(& (X)->lock); (X)->head=(X)->tail=0; }while(0)
 
-int co_wait(COND_T *c, MUTEX_T *m);
-int co_signal(COND_T *c);
-int co_broadcast(COND_T *c);
-int co_destroy(COND_T *c);
+PMOD_EXPORT int co_wait(COND_T *c, PIKE_MUTEX_T *m);
+PMOD_EXPORT int co_signal(COND_T *c);
+PMOD_EXPORT int co_broadcast(COND_T *c);
+PMOD_EXPORT int co_destroy(COND_T *c);
 
 #endif
 
+extern int th_running;
 
-extern MUTEX_T interpreter_lock;
+PMOD_EXPORT extern PIKE_MUTEX_T interpreter_lock;
 
-extern COND_T live_threads_change;		/* Used by _disable_threads */
-extern COND_T threads_disabled_change;		/* Used by _disable_threads */
+#if defined(PIKE_DEBUG) && !defined(__NT__)
+
+/* This is a debug wrapper to enable checks that the interpreter lock
+ * is hold by the current thread. */
+
+extern THREAD_T debug_locking_thread;
+#define SET_LOCKING_THREAD (debug_locking_thread = th_self(), 0)
+
+#define mt_lock_interpreter()						\
+  do {mt_lock(&interpreter_lock); SET_LOCKING_THREAD;} while (0)
+#define mt_trylock_interpreter() (mt_trylock(&interpreter_lock) || SET_LOCKING_THREAD)
+#define mt_unlock_interpreter() do {mt_unlock(&interpreter_lock);} while (0)
+#define co_wait_interpreter(COND) \
+  do {co_wait((COND), &interpreter_lock); SET_LOCKING_THREAD;} while (0)
+
+#define CHECK_INTERPRETER_LOCK() do {					\
+  if (th_running) {							\
+    THREAD_T self;							\
+    if (!mt_trylock(&interpreter_lock))					\
+      fatal("Interpreter is not locked.\n");				\
+    self = th_self();							\
+    if (!th_equal(debug_locking_thread, self))				\
+      fatal("Interpreter is not locked by this thread.\n");		\
+  }									\
+} while (0)
+
+#else
+
+#define mt_lock_interpreter() do {mt_lock(&interpreter_lock);} while (0)
+#define mt_trylock_interpreter() (mt_trylock(&interpreter_lock))
+#define mt_unlock_interpreter() do {mt_unlock(&interpreter_lock);} while (0)
+#define co_wait_interpreter(COND) do {co_wait((COND), &interpreter_lock);} while (0)
+
+#endif
+
+PMOD_EXPORT extern COND_T live_threads_change;		/* Used by _disable_threads */
+PMOD_EXPORT extern COND_T threads_disabled_change;		/* Used by _disable_threads */
 
 struct svalue;
-struct frame;
+struct pike_frame;
 
-extern MUTEX_T interleave_lock;
+extern PIKE_MUTEX_T interleave_lock;
 
 struct interleave_mutex
 {
   struct interleave_mutex *next;
   struct interleave_mutex *prev;
-  MUTEX_T lock;
+  PIKE_MUTEX_T lock;
 };
 
 #define IMUTEX_T struct interleave_mutex
@@ -281,35 +434,23 @@ struct interleave_mutex
 #define THREAD_EXITED 1
 
 struct thread_state {
+  struct Pike_interpreter state;
   char swapped;
   char status;
   COND_T status_change;
   THREAD_T id;
-  struct thread_state *hashlink, **backlink;
   struct mapping *thread_local;
-
-  /* Swapped variables */
-  struct svalue *sp,*evaluator_stack;
-  struct svalue **mark_sp,**mark_stack;
-  struct frame *fp;
-  int evaluator_stack_malloced;
-  int mark_stack_malloced;
-  JMP_BUF *recoveries;
-  struct object * thread_id;
-  char *stack_top;
-
+  struct thread_state *hashlink, **backlink;
 #ifdef PROFILING
-#ifdef HAVE_GETHRTIME
-  long long accounted_time;
+#if SIZEOF_LONG_LONG - 0 != 0
   long long time_base;
+#else
+  long time_base;
 #endif
-#endif
-
-#ifdef THREAD_TRACE
-  int t_flag;
-#endif /* THREAD_TRACE */
+#endif /* PROFILING */
 };
 
+
 #ifndef TH_RETURN_TYPE
 #define TH_RETURN_TYPE void *
 #endif
@@ -323,7 +464,12 @@ struct thread_state {
 #endif
 
 #ifndef th_yield
-#define th_yield()
+#ifdef HAVE_THR_YIELD
+#define th_yield() thr_yield()
+#else
+#define th_yield() 0
+#define HAVE_NO_YIELD
+#endif
 #endif
 
 #ifndef th_equal
@@ -334,6 +480,20 @@ struct thread_state {
 #define th_hash(X) hashmem((unsigned char *)&(X),sizeof(THREAD_T), 16)
 #endif
 
+#if !defined(HAVE_GETHRTIME) && \
+    !(defined(HAVE_MACH_TASK_INFO_H) && defined(TASK_THREAD_TIMES_INFO)) && \
+    defined(HAVE_CLOCK) && \
+    !defined(HAVE_NO_YIELD)
+#ifdef HAVE_TIME_H
+#include <time.h>
+#endif
+PMOD_EXPORT extern clock_t thread_start_clock;
+#define USE_CLOCK_FOR_SLICES
+#define DO_IF_USE_CLOCK_FOR_SLICES(X) X
+#else
+#define DO_IF_USE_CLOCK_FOR_SLICES(X)
+#endif
+
 /* Define to get a debug-trace of some of the threads operations. */
 /* #define VERBOSE_THREADS_DEBUG	0 */ /* Some debug */
 /* #define VERBOSE_THREADS_DEBUG	1 */ /* Lots of debug */
@@ -347,8 +507,9 @@ struct thread_state {
 #endif /* VERBOSE_THREADS_DEBUG */
 
 #ifdef THREAD_TRACE
-#define SWAP_OUT_TRACE(_tmp)	do { extern int t_flag; (_tmp)->t_flag = t_flag; } while(0)
-#define SWAP_IN_TRACE(_tmp)	do { extern int t_flag; t_flag = (_tmp)->t_flag; } while(0)
+PMOD_EXPORT extern int t_flag;
+#define SWAP_OUT_TRACE(_tmp)	do { (_tmp)->state.t_flag = t_flag; } while(0)
+#define SWAP_IN_TRACE(_tmp)	do { t_flag = (_tmp)->state.t_flag; } while(0)
 #else /* !THREAD_TRACE */
 #define SWAP_OUT_TRACE(_tmp)
 #define SWAP_IN_TRACE(_tmp)
@@ -360,49 +521,25 @@ struct thread_state {
 #define DO_IF_PROFILING(X)
 #endif
 
-#ifdef PIKE_DEBUG
-#define DO_IF_DEBUG(X) X
-#else
-#define DO_IF_DEBUG(X)
-#endif
-
-#define SWAP_OUT_THREAD(_tmp) do { \
-       (_tmp)->swapped=1; \
-       (_tmp)->evaluator_stack=evaluator_stack;\
-       (_tmp)->evaluator_stack_malloced=evaluator_stack_malloced;\
-       (_tmp)->fp=fp;\
-       (_tmp)->mark_sp=mark_sp;\
-       (_tmp)->mark_stack=mark_stack;\
-       (_tmp)->mark_stack_malloced=mark_stack_malloced;\
-       (_tmp)->recoveries=recoveries;\
-       (_tmp)->sp=sp; \
-       (_tmp)->stack_top=stack_top; \
-       (_tmp)->thread_id=thread_id;\
-       DO_IF_PROFILING( (_tmp)->accounted_time=accounted_time; ) \
-       DO_IF_PROFILING( (_tmp)->time_base = gethrtime() - time_base; ) \
-       SWAP_OUT_TRACE(_tmp); \
+#define SWAP_OUT_THREAD(_tmp) do {				\
+       (_tmp)->state=Pike_interpreter;				\
+       (_tmp)->swapped=1;					\
+       DO_IF_PROFILING( (_tmp)->time_base += gethrtime() ; )	\
+       SWAP_OUT_TRACE(_tmp);					\
+       Pike_interpreter.thread_id = (struct object *) (ptrdiff_t) -1;	\
       } while(0)
 
-#define SWAP_IN_THREAD(_tmp) do {\
-       (_tmp)->swapped=0; \
-       evaluator_stack=(_tmp)->evaluator_stack;\
-       evaluator_stack_malloced=(_tmp)->evaluator_stack_malloced;\
-       fp=(_tmp)->fp;\
-       mark_sp=(_tmp)->mark_sp;\
-       mark_stack=(_tmp)->mark_stack;\
-       mark_stack_malloced=(_tmp)->mark_stack_malloced;\
-       recoveries=(_tmp)->recoveries;\
-       sp=(_tmp)->sp;\
-       stack_top=(_tmp)->stack_top;\
-       thread_id=(_tmp)->thread_id;\
-       DO_IF_PROFILING( accounted_time=(_tmp)->accounted_time; ) \
-       DO_IF_PROFILING(  time_base =  gethrtime() - (_tmp)->time_base; ) \
-       SWAP_IN_TRACE(_tmp); \
+#define SWAP_IN_THREAD(_tmp) do {					\
+       (_tmp)->swapped=0;						\
+       Pike_interpreter=(_tmp)->state;					\
+       DO_IF_USE_CLOCK_FOR_SLICES (thread_start_clock = 0);		\
+       DO_IF_PROFILING(  Pike_interpreter.time_base -=  gethrtime();)	\
+       SWAP_IN_TRACE(_tmp);						\
      } while(0)
 
 #define SWAP_OUT_CURRENT_THREAD() \
   do {\
-     struct thread_state *_tmp=(struct thread_state *)thread_id->storage; \
+     struct thread_state *_tmp=OBJ2THREAD(Pike_interpreter.thread_id); \
      SWAP_OUT_THREAD(_tmp); \
      THREADS_FPRINTF(1, (stderr, "SWAP_OUT_CURRENT_THREAD() %s:%d t:%08x\n", \
 			 __FILE__, __LINE__, (unsigned int)_tmp->thread_id)) \
@@ -420,11 +557,8 @@ struct thread_state {
  * environment.
  */
 #define HIDE_GLOBAL_VARIABLES() do { \
-   int sp = 0, evaluator_stack = 0, mark_sp = 0, mark_stack = 0, fp = 0; \
-   void *evaluator_stack_malloced = NULL, *mark_stack_malloced = NULL; \
-   int recoveries = 0, thread_id = 0; \
-   int error = 0, xalloc = 0, low_my_putchar = 0, low_my_binary_strcat = 0; \
-   int low_make_buf_space = 0, pop_n_elems = 0; \
+   int Pike_interpreter =0; \
+   int pop_n_elems = 0; \
    int push_sp_mark = 0, pop_sp_mark = 0, threads_disabled = 1
 
 /* Note that the semi-colon below is needed to add an empty statement
@@ -435,51 +569,77 @@ struct thread_state {
 #define HIDE_GLOBAL_VARIABLES()
 #define REVEAL_GLOBAL_VARIABLES()
 #endif /* PIKE_DEBUG */
-			   
 
+#define	OBJ2THREAD(X) \
+  ((struct thread_state *)((X)->storage+thread_storage_offset))
+
+#define THREADSTATE2OBJ(X) ((X)->state.thread_id)
+
+PMOD_EXPORT extern int Pike_in_gc;
 #define THREADS_ALLOW() do { \
-     struct thread_state *_tmp=(struct thread_state *)thread_id->storage; \
+     struct thread_state *_tmp=OBJ2THREAD(Pike_interpreter.thread_id); \
+     DO_IF_DEBUG({ \
+       if(thread_for_id(th_self()) != Pike_interpreter.thread_id) \
+	 fatal("thread_for_id() (or Pike_interpreter.thread_id) failed!" \
+               " %p != %p\n", \
+               thread_for_id(th_self()), Pike_interpreter.thread_id); \
+       if (Pike_in_gc > 50 && Pike_in_gc < 300) \
+	 fatal("Threads allowed during garbage collection.\n"); \
+     }) \
      if(num_threads > 1 && !threads_disabled) { \
        SWAP_OUT_THREAD(_tmp); \
        THREADS_FPRINTF(1, (stderr, "THREADS_ALLOW() %s:%d t:%08x(#%d)\n", \
 			   __FILE__, __LINE__, \
 			   (unsigned int)_tmp->thread_id, live_threads)); \
-       mt_unlock(& interpreter_lock); \
+       mt_unlock_interpreter(); \
      } else {} \
      HIDE_GLOBAL_VARIABLES()
 
 #define THREADS_DISALLOW() \
      REVEAL_GLOBAL_VARIABLES(); \
      if(_tmp->swapped) { \
-       mt_lock(& interpreter_lock); \
+       mt_lock_interpreter(); \
        THREADS_FPRINTF(1, (stderr, "THREADS_DISALLOW() %s:%d t:%08x(#%d)\n", \
 			   __FILE__, __LINE__, \
                            (unsigned int)_tmp->thread_id, live_threads)); \
        while (threads_disabled) { \
          THREADS_FPRINTF(1, (stderr, \
                              "THREADS_DISALLOW(): Threads disabled\n")); \
-         co_wait(&threads_disabled_change, &interpreter_lock); \
+         co_wait_interpreter(&threads_disabled_change); \
        } \
        SWAP_IN_THREAD(_tmp);\
      } \
+     DO_IF_DEBUG( if(thread_for_id(th_self()) != Pike_interpreter.thread_id) \
+        fatal("thread_for_id() (or Pike_interpreter.thread_id) failed! %p != %p\n",thread_for_id(th_self()),Pike_interpreter.thread_id) ; ) \
    } while(0)
 
 #define THREADS_ALLOW_UID() do { \
-     struct thread_state *_tmp_uid=(struct thread_state *)thread_id->storage; \
+     struct thread_state *_tmp_uid=OBJ2THREAD(Pike_interpreter.thread_id); \
+     DO_IF_DEBUG({ \
+       if(thread_for_id(th_self()) != Pike_interpreter.thread_id) { \
+	 fatal("thread_for_id() (or Pike_interpreter.thread_id) failed! %p != %p\n", \
+               thread_for_id(th_self()),Pike_interpreter.thread_id); \
+       } \
+       if ((Pike_in_gc > 50) && (Pike_in_gc < 300)) { \
+         fprintf(stderr, __FILE__ ":" DEFINETOSTR(__LINE__) ": Fatal error:\n"); \
+	 debug_fatal("Threads allowed during garbage collection (%d).\n", \
+                     Pike_in_gc); \
+       } \
+     }) \
      if(num_threads > 1 && !threads_disabled) { \
        SWAP_OUT_THREAD(_tmp_uid); \
        live_threads++; \
        THREADS_FPRINTF(1, (stderr, "THREADS_ALLOW_UID() %s:%d t:%08x(#%d)\n", \
 			   __FILE__, __LINE__, \
 			   (unsigned int)_tmp_uid->thread_id, live_threads)); \
-       mt_unlock(& interpreter_lock); \
+       mt_unlock_interpreter(); \
      } else {} \
      HIDE_GLOBAL_VARIABLES()
 
 #define THREADS_DISALLOW_UID() \
      REVEAL_GLOBAL_VARIABLES(); \
      if(_tmp_uid->swapped) { \
-       mt_lock(& interpreter_lock); \
+       mt_lock_interpreter(); \
        live_threads--; \
        THREADS_FPRINTF(1, (stderr, \
                            "THREADS_DISALLOW_UID() %s:%d t:%08x(#%d)\n", \
@@ -488,22 +648,51 @@ struct thread_state {
        co_broadcast(&live_threads_change); \
        while (threads_disabled) { \
          THREADS_FPRINTF(1, (stderr, "THREADS_DISALLOW_UID(): Wait...\n")); \
-         co_wait(&threads_disabled_change, &interpreter_lock); \
+         co_wait_interpreter(&threads_disabled_change); \
        } \
        SWAP_IN_THREAD(_tmp_uid);\
      } \
    } while(0)
 
+#define SWAP_IN_THREAD_IF_REQUIRED() do { 			\
+  struct thread_state *_tmp=thread_state_for_id(th_self());	\
+  HIDE_GLOBAL_VARIABLES();					\
+  THREADS_DISALLOW()
+
+#ifdef PIKE_DEBUG
+#define ASSERT_THREAD_SWAPPED_IN() do {				\
+    struct thread_state *_tmp=thread_state_for_id(th_self());	\
+    if(_tmp->swapped) fatal("Thread is not swapped in!\n");	\
+  }while(0)
+
+#else
+#define ASSERT_THREAD_SWAPPED_IN()
+#endif
+
 /* Prototypes begin here */
-int low_nt_create_thread(unsigned stack_size,
+int low_nt_create_thread(unsigned Pike_stack_size,
 			 unsigned (TH_STDCALL *func)(void *),
 			 void *arg,
 			 unsigned *id);
 struct thread_starter;
+struct thread_local;
+void low_init_threads_disable(void);
+void init_threads_disable(struct object *o);
+void exit_threads_disable(struct object *o);
+void init_interleave_mutex(IMUTEX_T *im);
+void exit_interleave_mutex(IMUTEX_T *im);
+void thread_table_init(void);
+unsigned INT32 thread_table_hash(THREAD_T *tid);
+PMOD_EXPORT void thread_table_insert(struct object *o);
+PMOD_EXPORT void thread_table_delete(struct object *o);
+PMOD_EXPORT struct thread_state *thread_state_for_id(THREAD_T tid);
+PMOD_EXPORT struct object *thread_for_id(THREAD_T tid);
+PMOD_EXPORT void f_all_threads(INT32 args);
+PMOD_EXPORT int count_pike_threads(void);
 TH_RETURN_TYPE new_thread_func(void * data);
 void f_thread_create(INT32 args);
 void f_thread_set_concurrency(INT32 args);
-void f_this_thread(INT32 args);
+PMOD_EXPORT void f_this_thread(INT32 args);
 struct mutex_storage;
 struct key_storage;
 void f_mutex_lock(INT32 args);
@@ -521,27 +710,24 @@ void f_thread_backtrace(INT32 args);
 void f_thread_id_status(INT32 args);
 void init_thread_obj(struct object *o);
 void exit_thread_obj(struct object *o);
-void th_farm(void (*fun)(void *), void *);
-void th_init(void);
+PMOD_EXPORT void f_thread_local(INT32 args);
+void f_thread_local_get(INT32 args);
+void f_thread_local_set(INT32 args);
 void low_th_init(void);
+void th_init(void);
 void th_cleanup(void);
-void thread_table_insert(struct object *o);
-void thread_table_delete(struct object *o);
-struct thread_state *thread_state_for_id(THREAD_T tid);
-struct object *thread_for_id(THREAD_T tid);
-void f_all_threads(INT32 args);
-void f_thread_local(INT32 args);
-
-void low_init_threads_disable(void);
-void init_threads_disable(struct object *o);
-void exit_threads_disable(struct object *o);
-
-void init_interleave_mutex(IMUTEX_T *im);
-void exit_interleave_mutex(IMUTEX_T *im);
-
+int th_num_idle_farmers(void);
+int th_num_farmers(void);
+PMOD_EXPORT void th_farm(void (*fun)(void *), void *here);
 /* Prototypes end here */
 
 #else
+
+#define th_atfork(X,Y,Z)
+#define th_atfork_prepare()
+#define th_atfork_parent()
+#define th_atfork_child()
+
 #define th_setconcurrency(X)
 #define DEFINE_MUTEX(X)
 #define DEFINE_IMUTEX(X)
@@ -558,11 +744,14 @@ void exit_interleave_mutex(IMUTEX_T *im);
 #define THREADS_DISALLOW_UID()
 #define HIDE_GLOBAL_VARIABLES()
 #define REVEAL_GLOBAL_VARIABLES()
+#define ASSERT_THREAD_SWAPPED_IN()
+#define SWAP_IN_THREAD_IF_REQUIRED()
 #define th_init()
 #define low_th_init()
 #define th_cleanup()
 #define th_init_programs()
 #define th_self() ((void*)0)
+#define co_wait(X,Y)
 #define co_signal(X)
 #define co_broadcast(X)
 
@@ -570,16 +759,45 @@ void exit_interleave_mutex(IMUTEX_T *im);
 #define init_threads_disable(X)
 #define exit_threads_disable(X)
 
-#endif /* _REENTRANT */
+
+#endif /* PIKE_THREADS */
+
+#ifndef CHECK_INTERPRETER_LOCK
+#define CHECK_INTERPRETER_LOCK() do {} while (0)
+#endif
 
 #ifdef __NT__
 #ifndef PIKE_DEBUG
-#define CheckValidHandle(X) 0
+#define CheckValidHandle(X) (X)
 #else
-void CheckValidHandle(HANDLE h);
+PMOD_EXPORT HANDLE CheckValidHandle(HANDLE h);
 #endif
 #endif
 
 extern int threads_disabled;
+PMOD_EXPORT extern ptrdiff_t thread_storage_offset;
+PMOD_EXPORT extern struct program *thread_id_prog;
+
+#ifndef NO_PIKE_SHORTHAND
+#define MUTEX_T PIKE_MUTEX_T
+#endif
+
+
+/* Initializer macros for static mutex and condition variables */
+#ifdef PTHREAD_MUTEX_INITIALIZER
+#define STATIC_MUTEX_INIT  = PTHREAD_MUTEX_INITIALIZER
+#else
+#define STATIC_MUTEX_INIT
+#endif
+#ifdef PTHREAD_COND_INITIALIZER
+#define STATIC_COND_INIT   = PTHREAD_COND_INITIALIZER
+#else
+#define STATIC_COND_INIT
+#endif
+
+
+/* for compatibility */
+#include "interpret.h"
 
 #endif /* THREADS_H */
+
